// Description: generates events describing directory files
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//</ul>
//
// Category: System
// Author: Alex Hill
// Revision: 08/15/05 Alex Hill
//
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <dirent.h>
#else
#include <direct.h>
#include "dirent.h"
#endif
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include "directoryListing.h"

using namespace std;

extern "C" ygNode* APIENTRY construct_directoryListing(const char* name,bool master) { return new directoryListing(name,master); }

#ifndef WIN32
#define PATH_SEPARATOR ":"
#else
#define PATH_SEPARATOR ";"

static void ReverseSlashes(ygString& path)
    {
    char buffer[256];
    GetShortPathName(path.c_str(),buffer,256);
    path = buffer;
    char *pathData = path.c_str();
    for (int i=0; i<path.length(); i++)
        {
        if (pathData[i] == '\\')
            pathData[i] = '/';
        }
    return;
    }

#endif

static bool IsAbsolutePath(ygString& path)
    {
#ifdef WIN32
    ReverseSlashes(path);
#endif
    char *pathData = path.c_str();
    // "\xyz\foo.txt", "\\server\xyz\foo.txt"  
    if (pathData[0] == '/')
        return true;
#ifdef WIN32
    // "C:\xyz\foo.txt"
    else if (isalpha(pathData[0]) && pathData[1] == ':') 
        return true;
#endif
    else
        return false;
    }

directoryListing::directoryListing(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("directoryListing");
    dirIndex = 0;
    maximum = 100;
    truncate = true;
    ygDirIndex = -1;
    addNetMessage("directory",&dirName,YG_NET_STRING);
    addNetMessage("directory",&dirIndex,YG_NET_INT);
    addNetMessage("truncate",&truncate,YG_NET_BOOL);
    addNetMessage("maximum",&maximum,YG_NET_INT);
    addEvent("directory");
    addEvent("changed");
    if (getenv("YG_PATH"))
        {
        char buffer[256];
        getcwd(buffer,256);
        ygString localDirectory(buffer);
#ifdef WIN32
        ReverseSlashes(localDirectory);
#endif
        ygDirectories.push_back(localDirectory);
        ygString searchPath;
        searchPath = getenv("YG_PATH");
        ygString* dir;
        int i=0;
        ygString testDir;
        while (dir = searchPath.nextToken(PATH_SEPARATOR,&i))
            {
            testDir = *dir;
            delete dir;
            if (!IsAbsolutePath(testDir))
                {
                ygString tempDir(localDirectory);
                tempDir += "/";
                tempDir += testDir;
                testDir = tempDir;
                }
            ygDirectories.push_back(testDir);
            }
        }
    }
    
void directoryListing::reset(void)
    {
    lastDir.clear();
    lastIndex = -1;
    ygDirIndex = -1;
    //clear the directory name
    resetNetMessage("directory");
    //set ygpath to true
    resetNetMessage("truncate");
    //set the maximum printout to 100
    resetNetMessage("maximum");
    directories.clear();
    files.clear();
    ygNode::reset();
    }

void directoryListing::message(const ygMessage& msg)
    {
    //generate events with directory and file names
    if (msg == "directory")
        {
        checkDirectory();
        }
    //generate events with directory and file names
    else if (msg == "index")
        {
        if (msg.numArgs() > 0)
            {
            dirIndex = msg.intArg(0);
            checkDirectory();
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the flag to truncate file names included in the YG_PATH
    else if (msg == "truncate")
        {
        return;
        }
    //set the total number of directories and files to report
    else if (msg == "maximum")
        {
        return;
        }
    else
        ygNode::message(msg);
    }

#ifdef WIN32
#define S_ISDIR(mode) (((mode) & _S_IFDIR) == _S_IFDIR)
#define S_ISREG(mode) (((mode) & _S_IFREG) == _S_IFREG)
#endif

void directoryListing::checkDirectory(void)
    {
    if (dirName.length() > 0 && dirName != lastDir)
        {
        if (dirName == "." || dirName == "./")
            {
            char buffer[256];
            getcwd(buffer,256);
            dirName = buffer;
#ifdef WIN32
            ReverseSlashes(dirName);
#endif
            }
        else if (!IsAbsolutePath(dirName))
            {
            ygString tempName(dirName);
            char buffer[256];
            getcwd(buffer,256);
            dirName = buffer;
            dirName += "/";
            dirName += tempName;
            }
        struct stat statbuf;
        //if the directory is a valid one
        if (stat(dirName.c_str(),&statbuf) == 0 && S_ISDIR(statbuf.st_mode))
            {
            ygDirIndex = -1;
            int max=0;
            for (int i=0;i<ygDirectories.size();i++)
                {
                int len = ygDirectories[i].length();
                if (dirName.length() >= len && !(strncasecmp(dirName.c_str(),ygDirectories[i].c_str(),len)))
                    {
                    ygDirIndex = i;
                    max = len;
                    }
                }
		printf("ygDirIndex:%d\n",ygDirIndex);
            DIR* handle;
            handle = opendir(dirName.c_str());
            if (handle)
                {
                directories.clear();
                files.clear();
                //find parent directory
                int endindex;
                int len = dirName.length();
                endindex = dirName.find_last_of("/",len-1);
                if (len > 1 && endindex == len-1)
                    {
                    endindex = dirName.find_last_of("/",len-2);
                    }
                if (endindex == 0)
                    parentName = "/";
                else if (endindex > len-1)
                    parentName = ".";
                else
                    parentName.assign(dirName,0,endindex);
                dirent* entry;
                //discard the current directory handle
                readdir(handle);
                //discard the parent directory handle
                readdir(handle);
                int count;
                count = 0;
                //for each entry up to the maximum
                while((entry = readdir(handle)))
                    {
                    ygString fullName(dirName);
                    fullName += "/";
                    fullName += entry->d_name;
#ifdef WIN32
                    ReverseSlashes(fullName);
#endif
                    if (stat(fullName.c_str(),&statbuf) == 0)
                        {
                        if (S_ISDIR(statbuf.st_mode))
                            {
                            ygString currDir(entry->d_name);
                            vector<ygString>::iterator i;
                            //order the directories in alphabetical order
                            for (i=directories.begin();i!=directories.end();i++)
                                {
                                if (currDir < (*i))
                                    break;
                                }
                            //insert the new directory into the list
                            directories.insert(i,currDir);
                            count++;
                            }
                        else if ((S_ISREG(statbuf.st_mode)))
                            {
                            ygString currFile(entry->d_name);
                            vector<ygString>::iterator i;
                            //order the files in alphabetical order
                            for (i=files.begin();i!=files.end();i++)
                                {
                                if (currFile < (*i))
                                    break;
                                }
                            //insert the new file into the list
                            files.insert(i,currFile);
                            count++;
                            }
                        }
                    //closedir(handle);
                    }
                lastIndex = -1;
                }
            }
        }
    if (dirIndex != lastIndex)
        {
        int dirCount = directories.size();
        int fileCount = files.size();
        int dirMax = dirIndex + maximum;
        if (dirMax > dirCount + fileCount)
            dirMax = dirCount + fileCount;
        ygString args("total=");
        char index[32];
        sprintf(index,"%d",dirCount+fileCount);
        args += index;
        args += " parent=\"";
        args += parentName;
        args += "\" path=";
        ygString pathName;
        if (truncate && ygDirIndex >= 0)
            {
            if (dirName.length() > ygDirectories[ygDirIndex].length())
                {
                pathName = dirName;
                pathName.erase(0,ygDirectories[ygDirIndex].length()+1);
                pathName += "/";
                }
            }
        else
            {
            pathName = dirName;
            pathName += "/";
            }
        args += pathName;
        eventOccurred("changed",args);
        int num=0;
        for (int i=dirIndex;i<dirMax;i++)
            {
            ygString args("name=\"");
            if (i < dirCount)
                {
                args += directories[i];
                args += "\" path=\"";
                args += dirName;
                args += "/";
                args += directories[i];
                args += "\" type=directory";
                }
            else
                {
                args += files[i-dirCount];
                args += "\" path=\"";
                args += pathName;
                args += files[i-dirCount];
                args += "\" type=file";
               }
            args += " number=";
            sprintf(index,"%d",num);
            args += index;
            eventOccurred("directory",args);
            num++;
            }
        }
    lastDir = dirName;
    lastIndex = dirIndex;
    }
