// Description: a transform that reflects scene geometry about an axis
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> the default axis is the Y axis (0,1,0)
//<li> the scene node should neither be a transform nor have a parent transform
//<li> you must add the following to your .ygrc file in order to use this node:<br>
//<ul>
//<li> setenv YG_PRELOAD_CLASSES reflection
//</ul>
//</ul>
//
// Category: Geometry
// Author: Alex Hill
// Revision: 11/10/01
//
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include "reflection.h"


extern "C" ygNode* APIENTRY construct_reflection(const char* name,bool master) { return new reflection(name,master); }


reflection::reflection(const char* name,bool master) : ygTransform(name,master)
    {
    setClassName("reflection",true);
    pfnode()->setTravFuncs(PFTRAV_DRAW,preDrawFunc,postDrawFunc);
    reflection_.makeIdent();
    axis.set(0,1,0);
    setAxis(axis);
    dcs = new pfDCS();
    pfnode()->addChild(dcs);
    //distribute the nodeName string
    addNetMessage("addNode",&nodeNames,YG_NET_STRING_ADD);
    addNetMessage("removeNode",&nodeNames,YG_NET_STRING_DELETE);
    dontSendNetMessage("removeNode");
    dontSaveNetMessage("removeNode");
    //distribute the reflection axis
    addNetMessage("axis",&axis,YG_NET_VEC3);
    }

reflection::~reflection(void)
    {
    }

void reflection::reset(void)
    {
    //clear nodeNames string
    resetNetMessage("addNode");
    setNode(nodeNames);
    //reset the reflection axis to the Y axis (0,1,0)
    resetNetMessage("axis");
    setAxis(axis);
    ygTransform::reset();
    }

void reflection::message(const ygMessage& msg)
    {
    //set the name of the node to reflect
    if (msg == "node")
        {
        if (msg.numArgs() == 1)
            {
            nodeNames = msg.stringArg(0);
            setNode(nodeNames);
            netMessageChanged("addNode");
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //add the name of the node to reflect
    else if (msg == "addNode")
        {
        setNode(nodeNames);
        }
    //remove the name of the node to reflect
    else if (msg == "removeNode")
        {
        setNode(nodeNames);
    netMessageChanged("addNode");
        }
    //set the axis for reflection
    else if (msg == "axis")
        {
        setAxis(axis);
        }
    else
        ygTransform::message(msg);
    }

void reflection::setAxis(const pfVec3& val)
    {
    axis = val;
    if (axis[0] == 1.0)
        reflection_.setRow(0,-1,0,0,0);
    else
        reflection_.setRow(0,1,0,0,0);
    if (axis[1] == 1.0)
        reflection_.setRow(1,0,-1,0,0);
    else
        reflection_.setRow(1,0,1,0,0);
    if (axis[2] == 1.0)
        reflection_.setRow(2,0,0,-1,0);
    else
        reflection_.setRow(2,0,0,1,0);
    }
    
void reflection::app(void)
    {
    pfMatrix selfMatrix;
    getTransform(selfMatrix);
    pfMatrix inverseSelfMatrix;
    inverseSelfMatrix.invertOrtho(selfMatrix);
    pfMatrix matrix;
    matrix = inverseSelfMatrix * reflection_;
    dcs->setMat(matrix);
    ygTransform::app();
    }

void reflection::setNode(const ygString& name)
    {
    //delete all existing DCS children
    while (dcs->getNumChildren() > 0)
        dcs->removeChild(dcs->getChild(0));
    //set nodeNames to the input string
    nodeNames = name;
    ygString* nodeName;
    ygNode* node;
    int i=0;
    while (nodeName = nodeNames.nextToken(" ",&i))
        {
        //find the associated node in the database
        node = ygNodeDB::find(*nodeName);
        //if the node was found then add the associated Performer node as a child of the DCS
        if (node)
            dcs->addChild(node->pfnode());
        delete nodeName;
        }
    }

int reflection::preDrawFunc(pfTraverser* trav,void* data)
    {
    glFrontFace(GL_CW);
    return PFTRAV_CONT;
    }
    
int reflection::postDrawFunc(pfTraverser* trav,void* data)
    {
    glFrontFace(GL_CCW);
    return PFTRAV_CONT;
    }


