// Description: a space that generates events in the presence of a user and wand
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> for volume messages refer to the <a href="ygSpace.html">ygSpace</a> node
//<li> this node requires the user have a <a href="ygHead.html">ygHead</a> node
//<li> this node requires the user have a <a href="ygWand.html">ygWand</a> node
//<li> the position mode determines if the reported position is local to the object or relative to world coordinates
//<li> all wand events are also generated by this node (ex. button1)
//</ul>
//
// Category: Trigger
// Author: Dave Pape
//           11/01/01
//           02/01/06 Alex Hill - added position event arguments and inside event
// Revision: 07/01/06 Alex Hill - added orientation event arguments
//
#include <ygUser.h>
#include <ygWand.h>
#include <ygUtil.h>
#include <ygWorld.h>
#include <ygNetMessages.h>
#include "wandTrigger.h"

extern "C" ygNode* APIENTRY construct_wandTrigger(const char* name,bool master) { return new wandTrigger(name,master); }

wandTrigger::wandTrigger(const char* name,bool master) : ygSpace(name,master)
    {
    setClassName("wandTrigger");
    debugFlag("inside", &debugInside);
    lastFrame = 0;
    local = true;
    addNetMessage("local",&local,YG_NET_BOOL);
    addEvent("enter");
    addEvent("exit");
    addEvent("inside");
    }

wandTrigger::~wandTrigger(void)
    {
    }


void wandTrigger::reset(void)
    {
    //clear set of users previously inside
    prevInside.clear();
    prevInsideUser.clear();
    lastFrame = ygWorld::frameNumber();
    //set reporting mode to local
    local = true;
    //for each user in the world
    for (int i=0; i < ygWorld::world()->numUsers(); i++)
        {
        ygUser* user = ygWorld::world()->user(i);
        //for each associated wand
        for (int j=0; j < user->numWands(); j++)
            {
            if (user->wand(j) && user->wand(j)->isOfClass("ygWand"))
                {
                ygWand* wand = (ygWand*) (user->wand(j));
                const vector<ygString>& events = getEvents();
                const vector<ygString>& wandEvents = wand->getEvents();
                //for each wand event
                for (int k=0; k<wandEvents.size(); k++)
                    {
                    bool found = false;
                    for (int l=0; l<events.size(); l++)
                        {
                        if (events[l] == wandEvents[k])
                            found = true;
                        }
                    //if not found then add event
                    if (!found)
                        addEvent(wandEvents[k]);
                    }
                }
            }
        }
    ygSpace::reset();
    }


void wandTrigger::message(const ygMessage& msg)
    {
    //set position reporting mode
    if (msg == "local")
        {
        return;
        }
    else
        ygSpace::message(msg);
    }


void wandTrigger::app(void)
    {
    vector<ygWand*> inside;
    vector<ygUser*> insideUser;
    int i,j,f;
    if (lastFrame < (ygWorld::frameNumber()-1))
        {
        prevInside.clear();
        prevInsideUser.clear();
        }
    lastFrame = ygWorld::frameNumber();
    //for each user in the world
    for (i=0; i < ygWorld::world()->numUsers(); i++)
        {
        ygUser * user = ygWorld::world()->user(i);
        //for each associated wand
        for (j=0; j < user->numWands(); j++)
            {
            if (user->wand(j) && user->wand(j)->isOfClass("ygWand"))
                {
                ygWand* wand = (ygWand*) (user->wand(j));
                if (wand && wand->isOn())
                    {
                    //if wand is within the space then
                    pfMatrix mat,invmat;
                    getTransform(mat);
                    invmat.invertAff(mat);
                    wand->getTransform(mat);
                    mat*=invmat;
                    pfCoord coord;
                    mat.getOrthoCoord(&coord);
                    if (contains(coord.xyz))
                        {
                        if (debugInside)
                            cout << name() << ": " << user->name() << " with wand " << wand->name() << " is inside\n";
                        //add to set of wands inside
                        inside.push_back(wand);
                        insideUser.push_back(user);
                        ygString args("wand=");
                        args += wand->name();
                        args += " user=";
                        args += user->name();
                        //if wand was not previously inside generate enter event
                        if (!previouslyInside(wand))
                            {
                            //a wand and associated user have entered the space
                            eventOccurred("enter",args);
                            }
                        if (!local)
                            {
                            user->head()->getTransform(mat);
                            mat.getOrthoCoord(&coord);
                            }
                        char value[32];
                        args += " xpos=";
                        sprintf(value,"%f",coord.xyz[0]);
                        args += value;
                        args += " ypos=";
                        sprintf(value,"%f",coord.xyz[1]);
                        args += value;
                        args += " zpos=";
                        sprintf(value,"%f",coord.xyz[2]);
                        args += value;
                        args += " xori=";
                        sprintf(value,"%f",coord.hpr[1]);
                        args += value;
                        args += " yori=";
                        sprintf(value,"%f",coord.hpr[2]);
                        args += value;
                        args += " zori=";
                        sprintf(value,"%f",coord.hpr[0]);
                        args += value;
                        //a wand is inside the space
                        eventOccurred("inside",args);
                        //generate any flags associated with the wand
                        if (wand->numFlags() > 0)
                            {
                            for (f=0; f < wand->numFlags(); f++)
                                {
                                ygString event;
                                wand->getFlag(f,event);
                                //any wand events are generated while in the space
                                eventOccurred(event,args);
                                }
                            }
                        }
                    }
                }
            }
        }
    prevInside.swap(inside);
    prevInsideUser.swap(insideUser);
    //for each wand inside
    for (i=0; i < inside.size(); i++)
        {
        //if not previously inside then generate exit event
        if (!previouslyInside(inside[i]))
            {
            ygString args("wand=");
            args += inside[i]->name();
            if (insideUser[i])
                {
                args += " user=";
                args += insideUser[i]->name();
                }
            //a wand and associated user have exited the space
            eventOccurred("exit",args);
            }
        }
    ygSpace::app();
    }


bool wandTrigger::previouslyInside(const ygWand *w) const
    {
    int i;
    for (i=0; i < prevInside.size(); i++)
        if (w == prevInside[i])
            return true;
    return false;
    }
