// Author: Dave Pape
//           11/01/01
//           05/02/03 Alex Hill - added --viewer option for distributed applications
//           06/28/05 Alex Hill - removed world reset call because reset is now done by ygParser
//           02/01/06 Alex Hill - transfered all messages duplicated by ygWorld there to reduce redundency
// Revision: 07/01/06 Alex Hill - added logic to use CAVElib distribution
//
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <Performer/pf/pfChannel.h>
#include <Performer/pfdu.h>
#include <Performer/pfutil.h>
#include <Performer/pr/pfLight.h>
#include <Performer/pr/pfGeoState.h>
#include <pfcave.h>
#include <QUANTA/QUANTA.hxx>
#include "ygNet.h"
#include "ygParser.h"
#include "ygWorld.h"
#include "ygNodeDB.h"
#include "ygNet.h"
#include "ygNode.h"
#include "ygUtil.h"
#include "ygControl.h"
#include "ygUser.h"
#include "ygCAVEViewer.h"
#include "ygFileRequest.h"

#include "ygDLL.h"

#ifdef WIN32
#include <process.h>
#include "QUANTA\QUANTAinit.hxx"
#endif

using namespace std;

#ifdef WIN32
#define READ_BUFFER_SIZE 512

static HANDLE hStdin; 
static INPUT_RECORD irInBuf[READ_BUFFER_SIZE];
static DWORD cNumRead;
static char line[512] = "";
static DWORD count = 0;

static void initStdinMessage(void);
#endif

static void initGraphics(pfChannel *,void *);
static void checkNetMessages(void);
static void checkStdinMessages(void);
static void handleMessage(char *message);
static void addCAVEClasses(void);

static bool createView=true;
static bool exitFlag=false;

void printUsers(ygWorld *world)
    {
    int i,j;
    printf("\n%d users:\n",world->numUsers());
    for (i=0; i < world->numUsers(); i++)
        {
        ygUser *u = world->user(i);
        printf("  %s; ",u->name().c_str());
        printf("head=%p",u->head());
        if (u->head())
            printf("(%s)",u->head()->name().c_str());
        printf("; wands=");
        for (j=0; j < u->numWands(); j++)
            {
            printf(" %p",u->wand(j));
            if (u->wand(j))
                printf("(%s)",u->wand(j)->name().c_str());
            }
        printf("\n");
        }
    }


static void parseCommandLine(int argc,char **argv,ygWorld * world)
    {
    ygParser parser;
    int i;
    for (i=1; i < argc; i++)
        {
        if (!strncmp(argv[i],"--",2))
            {
            if (!strcasecmp(argv[i],"--noview"))
                createView = false;
            else if (!strcasecmp(argv[i],"--viewer"))
                {
                ygNode* node;
                node = ygNodeDB::find(argv[++i]);
                if (node && node->isOfClass("ygUser"))
                    {
                    ygUser* user = (ygUser*)node;
                    ((ygCAVEViewer*)world->viewer())->setUser(user);
                    }
                else
                    fprintf(stderr,"ERROR: (--viewer) user '%s' not found \n",argv[i]);
                }
            else if (!strcasecmp(argv[i],"--cavetracking"))
                {
                ygNode* node;
                node = ygNodeDB::find(argv[++i]);
                if (node && node->isOfClass("ygUser"))
                    {
                    ygUser* user = (ygUser*)node;
                    if (user)
                        user->setCAVETracking(true);
                    }
                else
                    fprintf(stderr,"ERROR: (--cavetracking) user '%s' not found \n",argv[i]);
                }	
            else
                fprintf(stderr,"WARNING: unknown command line "
                    "option '%s'\n",argv[i]);
            }
        else
            {
            parser.parseFile(argv[i],world);
            }
        }
    }


int main(int argc,char **argv)
    {
    ygWorld *world;
    ygNode *root;
#ifndef WIN32
    setlinebuf(stdin);
#else
	initStdinMessage();
	ygInit();//QUANTAinit();		// to make getimeofday work correctly
#endif
    ygPrePFInit();
    pfInit();
    pfCAVEConfig(&argc,argv,NULL);
    CAVESetOption(CAVE_PROJ_INCLUDENAVIGATION,1);
    ygPrePFConfig();
    addCAVEClasses();
    pfConfig();
    pfCAVEInitChannels();
    bool threaded = false;
    pfCAVEPreFrame();
    pfFrame();
    pfCAVEPostFrame();
    pfSync();
    //create ygWorld
    ygString worldName;
    char str[256];
    if (gethostname(str,sizeof(str)) == 0)
        worldName = str;
    else
        worldName = "ygdrasil";
    ygMakeValidNodeName(worldName);
    worldName += "_";
    sprintf(str,"%x",getpid()%1000);
    worldName += str;
    ygNodeDB::createValidName("world",worldName);
    world = (ygWorld*)ygNodeDB::create("world",worldName);
    pfCAVEMasterChan()->setScene(world->pfscene());
    //setup a secondary transparency bin that is rendered last
    pfCAVEMasterChan()->setBinSort(PFSORT_PATCHYFOG_BIN+1,PFSORT_BACK_TO_FRONT,NULL);
    pfCAVEMasterChan()->setBinOrder(PFSORT_PATCHYFOG_BIN+1,PFSORT_PATCHYFOG_BIN_ORDER+1);
    //set the global lighting model to 2 sided and localized
    pfGeoState* gstate = new pfGeoState;
    pfLightModel* lmodel = new pfLightModel;
    lmodel->setTwoSide(PF_ON);
    lmodel->setLocal(PF_ON);
    gstate->setAttr(PFSTATE_LIGHTMODEL,lmodel);
    world->pfscene()->setGState(gstate);
    //distributed nodes must wait for the master to load scene and upload
    int frames = 300;
    if (getenv("YG_NET_DOWNLOAD_DELAY"))
        frames = atoi(getenv("YG_NET_DOWNLOAD_DELAY"));
    if (!CAVEDistribMaster())
        ygDelayPFFrame(frames);
    //all distributed nodes download at the same time
    ygNet::init();
    //parse the downloaded database and load models
    root = ygNodeDB::find("root");
    //allow all distributed nodes to load scenes before connecting
    if (!CAVEDistribMaster())
        ygDelayPFFrame(10);
    ygNet::connect();
    //all distributed nodes wait a different time to add their local root node
    int uploadDelay = 10;
    if (getenv("YG_NET_UPLOAD_DELAY"))
        uploadDelay = atoi(getenv("YG_NET_UPLOAD_DELAY"));
    frames = uploadDelay*CAVEConfig->DistribID;
    if (!CAVEDistribMaster())
        ygDelayPFFrame(frames);
    ygNet::storeNode(world);
    //find root node
    if (root)
        {
        printf("Found existing root node\n");
        world->setRoot(root);
        }
    else
        {
        printf("Creating new root node\n");
        world->createRoot();
        }
    parseCommandLine(argc,argv,world);
    if (world->viewer())
        world->viewer()->reset();
    if (createView)
        {
        pfCAVEPreDrawFunc((pfChanFuncType)initGraphics);
        while ((!CAVEgetbutton(CAVE_ESCKEY)) && (!exitFlag) &&(!world->quit()))
            {
            pfCAVEPreFrame();
            world->frame();
            pfFrame();
            pfCAVEPostFrame();
            pfCAVEPreDrawFunc(NULL);
            checkNetMessages();
//#ifndef WIN32
            checkStdinMessages();
//#endif
            pfSync();
            if (!threaded)
                {
                ygNet::thread();
                threaded = true;
                }
            }
        }
    else
        {
        bool doSleep = false;
        if (getenv("YG_SERVER_DO_SLEEP"))
            doSleep = true;
        ygSetServerMode();
        while (!exitFlag)
            {
            pfCAVEPreFrame();
            world->frame();
            pfCAVEPostFrame();
            checkNetMessages();
            checkStdinMessages();
#ifndef WIN32
            if (doSleep)
                usleep(1000);
#else
			if (doSleep)
				Sleep(1);
#endif
            if (!threaded)
                {
                ygNet::thread();
                threaded = true;
                }
            }
        }
    delete world;
    ygHalt();
    pfCAVEHalt();
    pfuExitUtil();
    pfuExitInput();
    pfExit();
    return 0;
    }


static void initGraphics(pfChannel *chan,void *)
    {
    /* This is needed to keep it from crashing an Impact when the textures
      exceed texture memory */
    int i,size;
    static int first=1;
    if (first)
        {
        pfList *texList = pfuMakeSceneTexList(chan->getScene());
        if (getenv("YG_PF_SHOW_TEXTURES"))
            pfuDownloadTexList(texList, PFUTEX_SHOW);
        else
            pfuDownloadTexList(texList, PFUTEX_APPLY);
        for (i=0, size=0; i < texList->getNum(); i++)
            size += pfuGetTexSize((pfTexture *) texList->get(i));
        printf("TOTAL TEXTURE SIZE: %d bytes\n",size);
        first = 0;
        }
    }

#ifndef WIN32
static void checkStdinMessages(void)
    {
    fd_set readfds;
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    FD_ZERO(&readfds);
    FD_SET(0, &readfds);
    if (select(1,&readfds,NULL,NULL,&timeout))
        {
        if (FD_ISSET(0,&readfds))
            {
            char line[512];
            fgets(line,sizeof(line),stdin);
            if ((strlen(line)>0) && (line[strlen(line)-1] == '\n'))
                line[strlen(line)-1] = '\0';
            handleMessage(line);
            }
        }
    }
#else
static void initStdinMessage(void)
{
    DWORD fdwMode; 
 
    // Get the standard input handle. 
 
    hStdin = GetStdHandle(STD_INPUT_HANDLE); 
    if (hStdin == INVALID_HANDLE_VALUE) 
        printf("GetStdHandle error\n"); 
	SetConsoleMode(hStdin, ENABLE_ECHO_INPUT|ENABLE_LINE_INPUT); 
	FlushConsoleInputBuffer(hStdin);
}

static void checkStdinMessages(void)
{ 
	// Wait for the events. 
    if (!PeekConsoleInput(hStdin,      // input buffer handle 
					irInBuf,     // buffer to read into 
					READ_BUFFER_SIZE,         // size of read buffer 
					&cNumRead) ) // number of records read 
        printf("PeekConsoleInput error\n"); 

	if (cNumRead > 0) {
		ReadConsoleInput(hStdin, irInBuf, READ_BUFFER_SIZE, &cNumRead);
		if (irInBuf[0].EventType == KEY_EVENT && irInBuf[0].Event.KeyEvent.bKeyDown && irInBuf[0].Event.KeyEvent.uChar.AsciiChar != 0) {
			if (irInBuf[0].Event.KeyEvent.uChar.AsciiChar == 13) {
				line[count] = '\0';
				handleMessage(line);
				cout << endl;
				line[0] = '\0';
				count = 0;
			}
			else
			{
				line[count++] = irInBuf[0].Event.KeyEvent.uChar.AsciiChar;
				cout << irInBuf[0].Event.KeyEvent.uChar.AsciiChar << flush;
			}
		}
	}
}
#endif

static void checkNetMessages(void)
    {
    static QUANTAnet_udp_c *sock = NULL;
    if (!sock)
        {
        int port=1967;
        sock = new QUANTAnet_udp_c;
        if (getenv("YG_MESSAGE_PORT"))
            port = atoi(getenv("YG_MESSAGE_PORT"));
        sock->init(port);
        sock->makeNonBlocking();
        }
    char message[1024];
    while (sock->receive(message,sizeof(message)) > 0)
        {
        printf("received message '%s'\n",message);
        handleMessage(message);
        }
    }


static bool allWhiteSpace(char *s)
    {
    while (*s)
        if (!isspace(*s))
            return false;
        else
            s++;
    return true;
    }


static void handleMessage(char *message)
    {
//printf("got message (%s)\n",message);
    if (allWhiteSpace(message))
        return;
    ygMessage msg;
    msg.parseString(message,NULL,NULL);
    if (msg.recipientName().length() > 0)
        ygWorld::world()->scheduleMessage(msg);
    else if (msg == "texcheck")
        {
        ygNode *node = ygWorld::world()->root();
        if (msg.numArgs() > 0)
            node = ygNodeDB::find(msg.stringArg(0),false);
        if (node)
            {
            int i,size=0;
            pfList * texList = pfuMakeTexList(node->pfnode());
            for (i=0, size=0; i < texList->getNum(); i++)
                size += pfuGetTexSize((pfTexture *) texList->get(i));
            printf("%s has %d textures, total size = %d bytes\n",
                node->name().c_str(),i,size);
            }
        }
    else if (msg == "stats")
        {
        ygNode *node = ygWorld::world()->root();
        if (msg.numArgs() > 0)
            node = ygNodeDB::find(msg.stringArg(0),false);
        if (node)
            node->printTimingStats();
        }
    else if (msg == "netstats")
        ygNet::showStats();
    else if (msg == "netreset")
        ygNet::reset();
    else if (msg == "exit")
        exitFlag = true;
    else if (msg == "quit" || msg == "graph" || msg == "create" || msg == "delete" || msg == "load" || msg == "save")
        ygWorld::world()->parseMessage(msg);
    //else
        //printf("WARNING: unrecognized command\n");
    }

#include "ygDSOLoader.h"
#include "ygCAVENavigator.h"
#include "ygCAVEHead.h"
#include "ygCAVETracker.h"
#include "ygCAVEWand.h"


static void addCAVEClasses(void)
    {
    ygDSOLoader::addConstructor("ygCAVENavigator",construct_ygCAVENavigator);
    ygDSOLoader::addConstructor("CAVENavigator",construct_ygCAVENavigator);
    ygDSOLoader::addConstructor("ygCAVEHead",construct_ygCAVEHead);
    ygDSOLoader::addConstructor("CAVEHead",construct_ygCAVEHead);
    ygDSOLoader::addConstructor("ygCAVETracker",construct_ygCAVETracker);
    ygDSOLoader::addConstructor("CAVETracker",construct_ygCAVETracker);
    ygDSOLoader::addConstructor("ygCAVEWand",construct_ygCAVEWand);
    ygDSOLoader::addConstructor("CAVEWand",construct_ygCAVEWand);
    }
