// Description: a grid for snapping transformations
//
//<b>notes:</b>
//<ul>
//<li> all parent node messages are inherited
//<li> see reset method for default settings
//<li> uncrecognized messages sent to a ygGrid node will be filtered and sent to the current "node"
//</ul>
//
// Category: Transformation
// Author: Alex Hill
// Revision: 02/01/06
//
#include <Performer/pf/pfGroup.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pf/pfGeode.h>
#include <ygWorld.h>
#include <ygNetMessages.h>
#include <ygPFObjectCache.h>
#include <ygNodeDB.h>
#include <Performer/pfdu.h>
#include "ygGrid.h"
#ifdef WIN32
#include "rint.h"
#endif

#include "ygDLL.h"

using namespace std;

#define YG_AXIS_XY 904
#define YG_AXIS_YZ 905
#define YG_AXIS_XZ 906

#define MAX_VERTS 10000
#define MIN_SPACING 0.00001
#define MIN_SIZE 0.00001

extern "C" YG_DLL_API ygNode* APIENTRY construct_ygGrid(const char* name,bool master) { return new ygGrid(name,master); }

ygGrid::ygGrid(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("ygGrid",true);
    geode = NULL;
    spacing = 1.0;
    length = 100.0;
    snapGrid = 0.1;
    snapRotate = 5.0;
    axis = YG_AXIS_XY;
    draw = DRAW_ON;
    node = NULL;
    addNetMessage("spacing",&spacing,YG_NET_FLOAT);
    addNetMessage("length",&length,YG_NET_FLOAT);
    addNetMessage("snapGrid",&snapGrid,YG_NET_FLOAT);
    addNetMessage("snapRotate",&snapRotate,YG_NET_FLOAT);
    addNetMessage("axis",&axis,YG_NET_INT);
    addNetMessageMap("axis","xy",YG_AXIS_XY);
    addNetMessageMap("axis","yz",YG_AXIS_YZ);
    addNetMessageMap("axis","xz",YG_AXIS_XZ);
    addNetMessage("draw",&draw,YG_NET_INT);
    addNetMessageMap("draw","false",DRAW_OFF);
    addNetMessageMap("draw","true",DRAW_ON);
    addNetMessageMap("draw","local",DRAW_LOCAL);
    addNetMessageMap("draw","remote",DRAW_REMOTE);
    addNetMessage("node",&nodeName,YG_NET_NODE);
    dontSaveNetMessage("node");
    //create ygNode message filter
    filterMessages = " when addChild removeChild debug print reset event unreliableMessage reliableMessage resetTree edit selected ";
    //create grid geosets
    createGeometry();
    }

ygGrid::~ygGrid(void)
    {
    }


void ygGrid::reset(void)
    {
    resetNetMessage("spacing");
    resetNetMessage("length");
    resetNetMessage("snapGrid");
    resetNetMessage("snapRotate");
    resetNetMessage("axis");
    //set the draw flag to true
    resetNetMessage("draw");
    resetNetMessage("node");
    node = NULL;
    setDraw(draw);
    ygNode::reset();
    }


void ygGrid::message(const ygMessage& msg)
    {
    ygString paddedMessage = " ";
    paddedMessage += msg.message();
    paddedMessage += " ";
    //set the spacing between grid marks
    if (msg == "spacing")
        {
        if (spacing <= MIN_SPACING)
            spacing = MIN_SPACING;
        if (length/(2*spacing) > MAX_VERTS)
            spacing = length / (2 * MAX_VERTS);
        createGeometry();
        }
    //set the overall size of the grid
    else if (msg == "length")
        {
        if (length <= MIN_SIZE)
            length = MIN_SIZE;
        if (length/(2*spacing) > MAX_VERTS)
            length = spacing * 2 * MAX_VERTS;
        createGeometry();
        }
    //set the grid axis
    else if (msg == "axis")
        {
        createGeometry();
        }
    //set if the grid is drawn
    else if (msg == "draw")
        {
        setDraw(draw);
        }
    //set node for reference line and message filtering
    else if (msg == "node")
        {
        if (nodeName.length() > 0)
            {
            node = ygNodeDB::find(nodeName,false);
            if (!node)
                msg.error(name(),"(node not found)");
            }
        else
            node = NULL;
        }
    //set the scalar grid snap value
    else if (msg == "snapGrid")
        {
        return;
        }
    //set the rotational grid snap value
    else if (msg == "snapRotate")
        {
        return;
        }
    //filter all other messages
    else if (!filterMessages.contains(paddedMessage))
        {
        filterMessage(msg);
        }
    else
        ygNode::message(msg);
    }


void ygGrid::app(void)
    {
    updateReference();
    ygNode::app();
    }


void ygGrid::setDraw(int val)
    {
    bool* flag;
    draw = val;
    flag = getDebug("display");
    if (*flag || val == DRAW_ON || (isNetMaster() && val == DRAW_LOCAL) || (!isNetMaster() && val == DRAW_REMOTE))
        geode->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF,PF_SET);
    else
        geode->setTravMask(PFTRAV_DRAW,0,PFTRAV_SELF,PF_SET);
    }


void ygGrid::updateDebug(void)
    {
    if (debugSelected())
        debugColor()->set(YG_SELECT_COLOR);
    else
        debugColor()->set(YG_DEBUG_COLOR);
    if (debugDisplay())
        geode->setTravMask(PFTRAV_ISECT, YG_ISECT_ALLMASK | YG_ISECT_DEBUGMASK, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    else
        geode->setTravMask(PFTRAV_ISECT, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    if (debugDisplay() || draw == DRAW_ON || (isNetMaster() && draw == DRAW_LOCAL) || (!isNetMaster() && draw == DRAW_REMOTE))
        geode->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF,PF_SET);
    else
        geode->setTravMask(PFTRAV_DRAW,0,PFTRAV_SELF,PF_SET);
    netMessageChanged("display");
    }


void ygGrid::updateReference(void)
    {
    if (node && node->isOfClass("ygTransform"))
        {
        pfVec3 pos;
        pos = node->origin(this);
        float xpos = pos[0];
        float ypos = pos[1];
        float distance = pos[2];    
        if (axis == YG_AXIS_YZ)
            {
            xpos = -pos[2];
            distance = pos[0];    
            }
        else if (axis == YG_AXIS_XZ)
            {
            ypos = pos[2];
            distance = -pos[1];
            } 
        int numPoints;
        float floatPoints;
        floatPoints = fabs(distance)/spacing;
        numPoints = (int)floatPoints;
        if (floatPoints == (float)numPoints)
            numPoints = numPoints - 1;
        if (numPoints > 2*numVerts)
            numPoints = numVerts;
        int index = 0;
        reference[index++].set(xpos,ypos,0.0);
        reference[index++].set(xpos,ypos,distance);
        float interval = spacing;
        if (distance < 0)
            interval = -spacing;
        float zpos = 0.0;
        for (int x=0; x<numPoints; x++)
            {
            zpos += interval;
            reference[index++].set(xpos-snapGrid,ypos,zpos);
            reference[index++].set(xpos+snapGrid,ypos,zpos);
            reference[index++].set(xpos,ypos-snapGrid,zpos);
            reference[index++].set(xpos,ypos+snapGrid,zpos);
            }

        refGset->setNumPrims(2*numPoints+1);
        pfMatrix matrix;
        if (axis == YG_AXIS_YZ)
            {
            matrix.makeRot(90.0,0,1,0);
            pfdXformGSet(refGset,matrix);
            }
        else if (axis == YG_AXIS_XZ)
            {
            matrix.makeRot(90.0,1,0,0);
            pfdXformGSet(refGset,matrix);
            }
        }
    else
        refGset->setNumPrims(0);
    }


void ygGrid::filterMessage(const ygMessage& msg)
    {
    float filter = snapGrid;
    if (msg == "orientation")
        filter = snapRotate;
    if (node)
        {
        ygString msgStr;
        msgStr = msg.message();
        msgStr += "(";
        char val[32];
        float result;
        for (int i=0; i<msg.numArgs(); i++)
            {
            if (i > 0)
                msgStr += ",";
            result = filter*rint(msg.floatArg(i)/filter);
            sprintf(val,"%g",result);
            msgStr += val;
            }
        msgStr += ")";
        ygMessage newMsg;
        newMsg.parseString(msgStr,node);
        ygWorld::world()->scheduleMessage(newMsg);
        }
    }


void ygGrid::createGeometry(void)
    {
    if (geode)
        {
        reference = NULL;
        refGset = NULL;
        while (geode->getNumGSets() > 0)
            {
            pfGeoSet* gset = geode->getGSet(0);
            if (gset)
                {
                geode->removeGSet(gset);
                pfDelete(gset);
                }
            }
        }
    else
        {
        geode = new pfGeode;
	if (!geode)
            {
            cerr << "ERROR: grid::createGoemetry: failed to create pfGeode\n";
            return;
            }
        setDebugNode((pfGroup*)geode);
        }
    makeGridGeosets();
    }


void ygGrid::makeGridGeosets(void)
    {
    float floatVerts;
    floatVerts = length/(2*spacing);
    numVerts = (int)floatVerts;
    if (floatVerts == (float)numVerts)
        numVerts = numVerts - 1;
    //create center grid lines
    pfVec3* verts  = (pfVec3*) pfMalloc(4*sizeof(pfVec3), pfGetSharedArena());
    float width = length/2.0;
    verts[0].set(0.0,-width,0.0);
    verts[1].set(0.0,width,0.0);
    verts[2].set(-width,0.0,0.0);
    verts[3].set(width,0.0,0.0);
    pfGeoSet* gset = new pfGeoSet;
    gset->setPrimType(PFGS_LINES);
    gset->setLineWidth(2.0);
    gset->setNumPrims(2);
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);
    pfGeoState* gstate = new pfGeoState;
    gstate->setMode(PFSTATE_ENWIREFRAME,1);
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    pfMatrix matrix;
    if (axis == YG_AXIS_YZ)
        {
        matrix.makeRot(90.0,0,1,0);
        pfdXformGSet(gset,matrix);
        }
    else if (axis == YG_AXIS_XZ)
        {
        matrix.makeRot(90.0,1,0,0);
        pfdXformGSet(gset,matrix);
        }
    geode->addGSet(gset);
    //create interior grid lines
    verts  = (pfVec3*) pfMalloc(8*numVerts*sizeof(pfVec3), pfGetSharedArena());
    int index = 0;
    float position = 0.0;
    for (int x=0; x<numVerts; x++)
        {
        position += spacing;
        verts[index++].set(position,-width,0.0);
        verts[index++].set(position,width,0.0);
        verts[index++].set(-position,-width,0.0);
        verts[index++].set(-position,width,0.0);
        verts[index++].set(-width,position,0.0);
        verts[index++].set(width,position,0.0);
        verts[index++].set(-width,-position,0.0);
        verts[index++].set(width,-position,0.0);
        }
    gset = new pfGeoSet;
    gset->setPrimType(PFGS_LINES);
    gset->setLineWidth(1.0);
    gset->setNumPrims(4*numVerts);
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    if (axis == YG_AXIS_YZ)
        pfdXformGSet(gset,matrix);
    else if (axis == YG_AXIS_XZ)
        pfdXformGSet(gset,matrix);
    geode->addGSet(gset);
    //create reference line geode
    reference  = (pfVec3*) pfMalloc((8*numVerts+2)*sizeof(pfVec3), pfGetSharedArena());
    refGset = new pfGeoSet;
    refGset->setPrimType(PFGS_LINES);
    refGset->setNumPrims(0);
    refGset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, reference, NULL);
    refGset->setGState(gstate);
    refGset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    geode->addGSet(refGset);
    //create polygon for exterior lines and intersection testing
    verts = (pfVec3*) pfMalloc(4*sizeof(pfVec3), pfGetSharedArena());
    verts[0].set(-width,-width,0);
    verts[1].set(width,-width,0);
    verts[2].set(width,width,0);
    verts[3].set(-width,width,0);
    int *lengths = (int *) pfMalloc(sizeof(int),pfGetSharedArena());
    lengths[0] = 4;
    gset = new pfGeoSet;
    gset->setPrimType(PFGS_POLYS);
    gset->setNumPrims(1);
    gset->setPrimLengths(lengths);
    gset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, NULL);
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    if (axis == YG_AXIS_YZ || axis == YG_AXIS_XZ)
        pfdXformGSet(gset,matrix);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    geode->addGSet(gset);
    geode->setTravMask(PFTRAV_ISECT, 0, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    geode->setTravMask(PFTRAV_DRAW, 0xffffffff, PFTRAV_SELF|PFTRAV_DESCEND, PF_SET);
    }
